module GuessGame(
   numTries
 , State(word, hits)
 , misses
 , initialState
 , Result(..)
 , processGuess
 ) where

import System.Random(randomRIO)
import Data.Char(isAlpha, toLower)
import Data.List(sort)

-- Global constant: number of unsuccessful tries allowed.
numTries :: Int
numTries = 8

-- Utility function to count how often an element occurs in a list
count :: Eq alpha=> alpha-> [alpha]-> Int
count a = length . filter (a ==)

-- Utility function which picks a random element from a (non-empty) list
pickRandom :: [alpha]-> IO alpha
pickRandom [] = error "pickRandom: empty list"
pickRandom xs = do 
  i <- randomRIO (0, length xs - 1)
  return $ xs !! i

-- The game state
data State = St { word   :: String -- Zu ratendes Wort
                , hits   :: String -- Schon geratene Buchstaben
                , miss   :: String -- Falsch geratene Buchstaben
                }

misses :: State-> String
misses = sort . (filter isAlpha) . miss

-- Pick an initial state (in particular, a word to guess)
initialState :: [String]-> IO State
initialState dict = do
  secret <- pickRandom (filter ((5 <). length) dict)
  return $ St {word= map toLower secret, hits= "", miss= ""}

-- A Result: a hit or miss, repetition (already guessed), everything guessed, or too many tries
data Result = Miss | Hit | Repetition | GuessedIt | TooManyTries

-- Process a guess, return a result
processGuess :: Char-> State-> (Result, State)
processGuess c st =
  if c `elem` (miss st++ hits st) then (Repetition, st)
  else if count c (word st) == 0 then
    let st'= st {miss= c:miss st}
    in  if length (miss st') >= numTries then (TooManyTries, st') else (Miss, st')
  else let st' = st {hits= c: hits st}
    in  if all (\d-> d `elem` (hits st')) (word st') then (GuessedIt, st') else (Hit, st')

